/* Copyright (C) 2005 Hiroyuki OYAMA
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "ap_config.h"
#include "http_log.h"

#include "rast/db.h"
#include "apr_strings.h"

#include "mc_template.h"
#include "mc_encode.h"
#include "default_template.h"
#include <strings.h>

typedef struct search_rast_conf {
    const char *db_path;
    rast_db_t *db;
    rast_search_option_t *options;
    mc_template_t *result_view;
    mc_template_t *default_view;
    apr_table_t *alt_view;
    const char *sort_property;
    int sort_order;
} search_rast_conf_t;


module AP_MODULE_DECLARE_DATA search_rast_module;




/*
 * query string util func
 *
 */
static void plus_to_space(char *str)
{
    register int i;
    for (i = 0; str[i]; i++) {
        if (str[i] == '+')
            str[i] = ' ';
    }
}

apr_table_t *parse_query(request_rec *r)
{
    apr_table_t *query = apr_table_make(r->pool, 1);
    const char *args, *chunk;

    args = r->args ? r->args : "";
//    if (r->method_number == M_POST)
//        args = apr_pstrcat(r->pool, parse_by_body(r), "&", args, NULL);
    while (*args && (chunk = ap_getword(r->pool, &args, '&'))) {
        char *key, *value;

        key   = ap_getword(r->pool, &chunk, '=');
        value = ap_getword(r->pool, &chunk, '=');
        plus_to_space(key);
        plus_to_space(value);
        ap_unescape_url(key);
        ap_unescape_url(value);
        apr_table_set(query, key, value);
    }

    return query;
}


apr_status_t cleanup_rast(void *dummy)
{
    rast_finalize();
    return APR_SUCCESS;
}

apr_status_t cleanup_rast_db(void *data)
{
    rast_db_t *db = data;

    if (db != NULL) {
        rast_db_close(db);
    }

    return APR_SUCCESS;
}

static int pre_config(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptmp)
{
    rast_error_t *error;

    error = rast_initialize();
    if (error != RAST_OK) {
        ap_log_error(APLOG_MARK, APLOG_STARTUP|APLOG_ERR, 0, NULL,
                     "mod_search_rast: %s\n", error->message);
        rast_error_destroy(error);
        return HTTP_INTERNAL_SERVER_ERROR;
    }
    apr_pool_cleanup_register(p, NULL, apr_pool_cleanup_null, cleanup_rast);

    return OK;
}

static int send_default_page(request_rec *r, const char *charset, apr_table_t *param)
{
    search_rast_conf_t *conf = ap_get_module_config(r->per_dir_config, &search_rast_module);
    const char *content;

    content = mc_template_build(r->pool, conf->default_view, param);
    r->content_type = apr_pstrcat(r->pool, "text/html; charset=", charset, NULL);
    ap_set_content_length(r, strlen(content));
    ap_rputs(content, r);

    return OK;
}


static rast_error_t *open_rast_db(apr_pool_t *p,
                                  search_rast_conf_t *conf, rast_db_t **db)
{
    rast_error_t *error;

    error = rast_db_open(db, conf->db_path, RAST_DB_RDONLY, NULL, p);
    if (error != RAST_OK) {
        return error;
    }

    apr_pool_cleanup_register(p, *db, cleanup_rast_db, cleanup_rast_db);

    return RAST_OK;
}

static int get_int_param(apr_table_t *param, const char *name)
{
    const char *value;
    char *endptr;
    long ret;

    value = apr_table_get(param, name);
    if (value == NULL)
        return 0;
    ret = strtol(value, &endptr, 10);
    return (int)ret;
}

static const char *get_db_encoding(rast_db_t *db)
{
    const char *db_encoding;

    db_encoding = rast_db_encoding(db);
    if (db_encoding == NULL) {
        return "UTF-8";
    }
    else if (strcmp(db_encoding, "euc_jp") == 0) {
        return "EUC-JP";
    }
    else if (strcmp(db_encoding, "utf8") == 0) {
        return "UTF-8";
    }
    else {
        return "UTF-8";
    }
}

static int search_rast_handler(request_rec *r)
{
    search_rast_conf_t *conf;
    rast_db_t *db;
    rast_search_option_t *option;
    apr_table_t *query;
    const char *keyword, *body, *type, *in_charset, *charset;
    rast_result_t *result;
    rast_error_t *error;
    apr_table_t *param, *page;
    apr_array_header_t *items, *pages;
    int i, per_page, page_num, start_no;
    mc_template_t *view;

    if (strcmp(r->handler, "search-rast"))
        return DECLINED;
    conf = ap_get_module_config(r->per_dir_config, &search_rast_module);

    error = open_rast_db(r->pool, conf, &db); 
    if (error != RAST_OK) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "mod_search_rast: cannot open database: %s", error->message);
        rast_error_destroy(error);
        return HTTP_INTERNAL_SERVER_ERROR;
    }
    charset = get_db_encoding(db);

    param = apr_table_make(r->pool, 5);
    query = parse_query(r);
    type = apr_table_get(query, "type");
    if (type == NULL || *type == '\0')
        type = "";
    keyword = apr_table_get(query, "q");
    if (keyword == NULL || *keyword == '\0') {
        return send_default_page(r, charset, param);
    }
    in_charset = apr_table_get(query, "ie");
    if (in_charset != NULL && *in_charset != '\0') {
        const char *org = keyword;
        keyword = mc_encode_convert(r->pool, in_charset, charset, keyword);
        if (keyword == NULL) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                          "mod_search_rast: cannot convert keyword '%s' to '%s'",
                          in_charset, charset);
            keyword = org;
        }
    }

    start_no = 0;
    per_page = get_int_param(query, "num");
    per_page = per_page > 0 ? per_page : 50;
    page_num = get_int_param(query, "page");
    if (page_num > 0) {
        start_no = page_num * per_page - per_page;
    }
    else {
        start_no = get_int_param(query, "start");
        page_num = (int)((start_no - 1)/per_page) + 1;   
    }

    option = rast_search_option_create(r->pool);
    option->start_no = start_no;
    option->num_items = 5000;
    option->need_summary = 1;
    option->summary_nchars = 200;
    option->properties = conf->options->properties;
    option->num_properties = conf->options->num_properties;
    if (conf->sort_property) {
        option->sort_method   = RAST_SORT_METHOD_PROPERTY;
        option->sort_property = conf->sort_property;
        option->sort_order    = conf->sort_order;
    }

    error = rast_db_search(db, keyword, option, &result, r->pool);    
    if (error != RAST_OK) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "mod_search_rast: cannot searcy keyword '%s': %s",
                      keyword, error->message);
        apr_table_set(param, "error", error->message);
        rast_error_destroy(error);
        return send_default_page(r, charset, param);
    }


    apr_table_setn(param, "num_per_page", apr_psprintf(r->pool, "%d", per_page));
    apr_table_setn(param, "page_num", apr_psprintf(r->pool, "%d", page_num));
    apr_table_setn(param, "keyword", keyword);
    apr_table_setn(param, "num_item", apr_psprintf(r->pool, "%d", result->hit_count));
    apr_table_setn(param, "type", type);
    apr_table_setn(param, "ie", in_charset ? in_charset : "");
    if (option->start_no < result->hit_count) {
        items = apr_array_make(r->pool, 0, sizeof(apr_table_t *));
        for (i = 0; i < result->num_items; i++) {
            apr_table_t *item = apr_table_make(r->pool, 3);
            int p;

            for (p = 0; p < conf->options->num_properties; p++) {
                rast_value_t *val = result->items[i]->properties + p;
                const char *name = conf->options->properties[p];
                if (val->type == RAST_TYPE_STRING) {
                    apr_table_set(item, name, rast_value_string(val));
                }
                else if (val->type == RAST_TYPE_UINT) {
                    apr_table_setn(item, name,
                                   apr_psprintf(r->pool,
                                                "%u", rast_value_uint(val)));
                }
                else if (val->type == RAST_TYPE_DATE) {
                    apr_table_set(item, name, rast_value_date(val));
                    /* apr_date_checkmast(date, "####-##-##T##:##:##"); */
                }
            }
            apr_table_setn(item, "summary",
                           apr_pstrndup(r->pool, result->items[i]->summary,
                                        result->items[i]->summary_nbytes));
            *(apr_table_t **)apr_array_push(items) = item;
            if (per_page <= i + 1)
                break;
        }
        apr_table_setn(param, "result", (const char *)items);
    }

    if (result->hit_count > per_page && option->start_no <= result->hit_count) {
        pages = apr_array_make(r->pool, 0, sizeof(apr_table_t *));
        for (i = 0; i * per_page < result->hit_count; i++) {
            page = apr_table_make(r->pool, 3);
            apr_table_set(page, "offset", apr_psprintf(r->pool, "%d", i * per_page));
            apr_table_set(page, "page_num", apr_psprintf(r->pool, "%d", i + 1));
            if (option->start_no >= i * per_page
                && option->start_no < (i + 1) * per_page)
            {
                apr_table_set(page, "current_page", "1");
            }
            *(apr_table_t **)apr_array_push(pages) = page;
        }
        apr_table_setn(param, "pages", (const char *)pages);
    }

    view = (mc_template_t *)apr_table_get(conf->alt_view, type);
    if (view == NULL) { /* use default view */
        body = mc_template_build(r->pool, conf->result_view, param);
        r->content_type = apr_pstrcat(r->pool, "text/html; charset=", charset, NULL);
    }
    else {
        char *view_charset;
        body = mc_template_build(r->pool, view, param);
        if ((view_charset = strcasestr(view->content_type, "; charset=")) != NULL) {
            const char *org = body;
            view_charset += strlen("; charset=");
            body = mc_encode_convert(r->pool, charset, view_charset, body);
            if (body == NULL) {
                body = org;
                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                              "mod_search_rast: cannot convert view charset '%s' to '%s'",
                              charset, view_charset);
            }
        }
        r->content_type = view->content_type;
    }

    ap_set_content_length(r, strlen(body));
    if (r->header_only) {
        return OK;
    }
    ap_rputs(body, r);

    return OK;
}


/*
 * Configuration directives
 *
 */
static void *create_dir_config(apr_pool_t *p, char *dir)
{
    search_rast_conf_t *conf;

    conf = apr_palloc(p, sizeof(search_rast_conf_t));
    conf->db_path = NULL;
    conf->db = NULL;
    conf->options = rast_search_option_create(p);
    conf->options->need_summary = 1;
    conf->result_view = mc_template_create_by_string(p, DEFAULT_RESULT_TEMPLATE);
    conf->default_view = mc_template_create_by_string(p, DEFAULT_DEFAULT_TEMPLATE);
    conf->alt_view = apr_table_make(p, 0);
    conf->sort_property = NULL;
    conf->sort_order = RAST_SORT_ORDER_DEFAULT;

    return (void *)conf;
}


static const char *set_and_open_rast_db(cmd_parms *cmd, void *dconf,
                                        const char *path)
{
    search_rast_conf_t *conf = dconf;
/*    rast_db_t *db; */
/*    rast_error_t *error; */

    if (conf->db != NULL) {
        return NULL;
    }
    conf->db_path = apr_pstrdup(cmd->pool, path);

/*
    error = rast_db_open(&db, path, RAST_RDONLY, NULL, cmd->pool);
    if (error != RAST_OK) {
        const char *msg;

        msg = apr_psprintf(cmd->pool,
                           "RastDatabase: (%d) cannot open database: %s: %s",
                           error->code, path, error->message);
        rast_error_destroy(error);
        return msg;
    }

    conf->db = db;
    apr_pool_cleanup_register(cmd->pool, db, cleanup_rast_db, cleanup_rast_db);
*/

    return NULL;
}


static const char *set_properties(cmd_parms *cmd, void *dconf, const char *name)
{
    search_rast_conf_t *conf = dconf;
    rast_search_option_t *options = conf->options;
    const char **properties, **new_properties;
    int i;

    if (options->properties == NULL) {
        new_properties = apr_palloc(cmd->pool, sizeof(const char *));
        new_properties[0] = apr_pstrdup(cmd->pool, name);
        options->properties = new_properties;
    }
    else {
        properties = options->properties;
        new_properties = apr_palloc(cmd->pool,
                                    sizeof(const char *)
                                        * (options->num_properties + 1));
        for (i  = 0; i < options->num_properties; i++) {
            new_properties[i] = properties[i];
        }
        new_properties[i] = apr_pstrdup(cmd->pool, name);
        options->properties = new_properties;
    }

    options->num_properties++;

    return NULL;
}


static const char *set_result_template(cmd_parms *cmd, void *dconf,
                                       const char *path)
{
    search_rast_conf_t *conf = dconf;

    conf->result_view = mc_template_create(cmd->pool, path);
    if (conf->result_view == NULL) {
        return "RastSetResultView: cannot open or create template file";
    }

    return NULL;
}

static const char *set_default_template(cmd_parms *cmd, void *dconf,
                                        const char *path)
{
    search_rast_conf_t *conf = dconf;

    conf->default_view = mc_template_create(cmd->pool, path);
    if (conf->default_view == NULL) {
        return "RastSetDefaultView: cannot open or create template file";
    }

    return NULL;
}

static const char *add_result_view(cmd_parms *cmd, void *dconf,
                                   const char *id, const char *path, const char *type)
{
    search_rast_conf_t *conf = dconf;
    mc_template_t *tmpl;

    tmpl = mc_template_create(cmd->pool, path);
    if (tmpl == NULL) {
        return "RastAddView: cannot open or create template file";
    }
    tmpl->content_type = type;
    apr_table_setn(conf->alt_view, id, (const char *)tmpl);

    return NULL;
}


static const char *set_sort_property(cmd_parms *cmd, void *dconf,
                                      const char *name, const char *order)
{
    search_rast_conf_t *conf = dconf;

    conf->sort_property = name;
    if (order != NULL && strcasecmp(order, "DESC") == 0) {
        conf->sort_order = RAST_SORT_ORDER_DESCENDING; 
    }
    else {
        conf->sort_order = RAST_SORT_ORDER_ASCENDING;
    }

    return NULL;
}

static const command_rec commands[] = {
    AP_INIT_TAKE1("RastDatabase", set_and_open_rast_db,
                  NULL,
                  ACCESS_CONF|OR_NONE, "set Rast database path"),
    AP_INIT_ITERATE("RastProperties", set_properties,
                    NULL,
                    ACCESS_CONF|OR_NONE, "set Rast search properties"),
    AP_INIT_TAKE1("RastSetResultView", set_result_template,
                  NULL,
                  ACCESS_CONF|OR_NONE, "set Rast search result page's template."),
    AP_INIT_TAKE1("RastSetDefaultView", set_default_template,
                  NULL,
                  ACCESS_CONF|OR_NONE, "set Rast default page's template."),
    AP_INIT_TAKE12("RastSortProperty", set_sort_property,
                  NULL,
                  ACCESS_CONF|OR_NONE, "set property name for search result sort"),
    AP_INIT_TAKE3("RastAddView", add_result_view,
                  NULL,
                  ACCESS_CONF|OR_NONE, "set Rast search result page's templates."),
    { NULL }
};


static void search_rast_register_hooks(apr_pool_t *p)
{
    ap_hook_handler(search_rast_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_pre_config(pre_config, NULL, NULL, APR_HOOK_MIDDLE);
}

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA search_rast_module = {
    STANDARD20_MODULE_STUFF, 
    create_dir_config,     /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    commands,              /* table of config file commands       */
    search_rast_register_hooks  /* register hooks                      */
};

